/*
 * File:        sidlObjA.c
 * Package:     sidl Python Object Adaptor
 * Copyright:   (c) 2001 The Regents of the University of California
 * Release:     $Name: release-0-10-0 $
 * Revision:    @(#) $Revision: 1.2 $
 * Date:        $Date: 2004/09/17 15:04:07 $
 * Description: Python extension type written in C for sidl object/interface
 *
 * This is a Python extension type written in C to wrap instances of
 * sidl objects or interfaces in a Python object.  If this looks
 * mysterious to you, look at Programming Python and the BABEL
 * documentation. 
 * Copyright (c) 2000-2001, The Regents of the University of Calfornia.
 * Produced at the Lawrence Livermore National Laboratory.
 * Written by the Components Team <components@llnl.gov>
 * UCRL-CODE-2002-054
 * All rights reserved.
 * 
 * This file is part of Babel. For more information, see
 * http://www.llnl.gov/CASC/components/. Please read the COPYRIGHT file
 * for Our Notice and the LICENSE file for the GNU Lesser General Public
 * License.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License (as published by
 * the Free Software Foundation) version 2.1 dated February 1999.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
 * conditions of the GNU Lesser General Public License for more details.
 * 
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#define sidlOBJA_MODULE 1
#include "sidlObjA.h"
#include "sidl_BaseClass_IOR.h"
#include "sidl_BaseException.h"

#include <string.h>

/**
 * This defines the Python object wrapper for a sidl object or interface.
 * This one Python extension type written in C can support an arbitrary
 * class or interface. The peculiarities of a particular class or interface
 * are stored in the d_methods member which defines what the object
 * can do.
 */
struct sidlPythonObject {
  PyObject_HEAD                 /* standard Python object header */

  /* sidl specific extensions */

  PyMethodDef                         *d_methods;
  struct sidl_BaseClass__object *    (*d_getIOR)(struct sidlPythonObject *);
  union {
    struct sidl_BaseClass__object     *d_ior;
    struct sidl_BaseInterface__object *d_abint;
  }                                    d_sidlobj;
  int				       d_refType;
};

/**
 * This <code>typedef</code> is required by PyObject_NEW; otherwise,
 * it is unused.
 */
typedef struct sidlPythonObject SPObject;

/**
 * A function to get the independent object representation (IOR) for
 * a wrapped instance of a class.
 */
static struct sidl_BaseClass__object *
sidlPython_Object_getIOR(struct sidlPythonObject *sobj) {
  return sobj->d_sidlobj.d_ior;
}

/**
 * A function to get the independent object representation (IOR) for
 * a wrapped instance of a interface.
 */
static struct sidl_BaseClass__object *
sidlPython_Interface_getIOR(struct sidlPythonObject *sobj) {
  return sobj->d_sidlobj.d_abint->d_object;
}

/**
 * Provide a forward declaration of the Python type object used
 * by <code>sidlPythonObject</code>'s.
 */
staticforward PyTypeObject sidlPythonObjectType;

#define is_SIDLobject(v) ((v)->ob_type == &sidlPythonObjectType)

/*
 * Exported C API methods
 */

static sidl_Object_Create_RETURN
sidl_Object_Create sidl_Object_Create_PROTO {
  PyObject *result = NULL;
  if (methods) {
    if (ior) {
      struct sidlPythonObject *self = 
	PyObject_NEW(SPObject, &sidlPythonObjectType);
      if (self){
	self->d_sidlobj.d_ior = ior;
        if (refType == sidl_PyNewRef) {
          (*ior->d_epv->f_addRef)(ior);
        }
	self->d_methods = methods;
	self->d_getIOR = sidlPython_Object_getIOR;
        self->d_refType = refType;
	result = (PyObject *)self;
      }
      else {
        if (refType == sidl_PyStealRef) {
          (*(ior->d_epv->f_deleteRef))(ior);
        }
      }
    }
    else {
      result = Py_None;
      Py_INCREF(result);
    }
  }
  else {
    PyErr_SetString(PyExc_AssertionError,
		    "sidl object has NULL IOR or vtable");
  }
  return result;
}

static sidl_Get_IOR_RETURN
sidl_Get_IOR sidl_Get_IOR_PROTO {
  if (is_SIDLobject(obj)) {
    return (*((struct sidlPythonObject *)obj)->d_getIOR)
      ((struct sidlPythonObject *)obj);
  }
  return NULL;
}

static sidl_Cast_RETURN
sidl_Cast sidl_Cast_PROTO {
  struct sidl_BaseClass__object *ior = sidl_Get_IOR(obj);
  return (ior) ? ((*(ior->d_epv->f__cast))(ior, name)) : NULL;
}

static sidl_Opaque_Create_RETURN
sidl_Opaque_Create sidl_Opaque_Create_PROTO {
  return PyCObject_FromVoidPtr(opaque_ptr, NULL);
}

static sidl_Opaque_Convert_RETURN
sidl_Opaque_Convert sidl_Opaque_Convert_PROTO {
  if (PyCObject_Check(obj)) {
    *opaque_ptr = PyCObject_AsVoidPtr(obj);
    return 1;
  }
  return 0;
}

static sidl_PyType_RETURN
sidl_PyType sidl_PyType_PROTO {
  return (PyTypeObject *)&sidlPythonObjectType;
}

static sidl_Interface_Create_RETURN
sidl_Interface_Create sidl_Interface_Create_PROTO {
  PyObject *result = NULL;
  if (methods) {
    if (abint) {
      struct sidlPythonObject *self = 
	PyObject_NEW(SPObject, &sidlPythonObjectType);
      if (self){
	self->d_sidlobj.d_abint = abint;
        if (refType == sidl_PyNewRef) {
          (*abint->d_epv->f_addRef)(abint->d_object);
        }
	self->d_methods = methods;
	self->d_getIOR = sidlPython_Interface_getIOR;
        self->d_refType = refType;
	result = (PyObject *)self;
      }
      else {
        if (refType == sidl_PyStealRef) {
          (*abint->d_epv->f_deleteRef)(abint->d_object);
        }
      }
    }
    else {
      result = Py_None;
      Py_INCREF(result);
    }
  }
  else {
    PyErr_SetString(PyExc_AssertionError,
		    "sidl interface has NULL vtable");
  }
  return result;
}

static sidl_Interface_Check_RETURN
sidl_Interface_Check sidl_Interface_Check_PROTO {
  if (is_SIDLobject(obj) && 
      (((struct sidlPythonObject *)obj)->d_getIOR ==
       sidlPython_Interface_getIOR)) {
    return ((struct sidlPythonObject *)obj)->d_sidlobj.d_abint;
  }
  return NULL;
}

static sidl_PyExceptionCast_RETURN
sidl_PyExceptionCast sidl_PyExceptionCast_PROTO {
  return sidl_BaseInterface__cast2(ex, name);
}

/*
 * BASIC OBJECT METHODS
 */

static PyObject *
spoa_getattr(struct sidlPythonObject *self,
            char *name)
{
  if (!strcmp(name, "__members__")){
    return Py_BuildValue("[]");
  }
  return Py_FindMethod(self->d_methods, (PyObject *)self, name);
}

static void
spoa_self_destruct(struct sidlPythonObject *self)
{
  struct sidl_BaseClass__object *ior = sidl_Get_IOR((PyObject *)self);
  /* remove Python's reference to this sidl object */
  if (self->d_refType != sidl_PyWeakRef) {
    (*(ior->d_epv->f_deleteRef))(ior);
  }
  self->d_sidlobj.d_ior = NULL;
  self->d_sidlobj.d_abint = NULL;
  self->d_getIOR = NULL;
  self->d_methods = NULL;
  self->d_refType = sidl_PyWeakRef;
  PyMem_DEL(self);
}

static PyTypeObject sidlPythonObjectType = {
  /* type header */
  PyObject_HEAD_INIT(NULL)
  0,
  "sidlObjA",
  sizeof(struct sidlPythonObject),
  0,

  /* standard methods */
  (destructor) spoa_self_destruct, /* call when reference count == 0 */
  (printfunc) 0,
  (getattrfunc) spoa_getattr,
  (setattrfunc) 0,
  (cmpfunc) 0,
  (reprfunc) 0,
 
  /* type categories */
  (PyNumberMethods *) 0,
  (PySequenceMethods *) 0,
  (PyMappingMethods *) 0,

  /* more methods */
  (hashfunc)   0,
  (ternaryfunc) 0,
  (reprfunc)   0, /* to string */
  (getattrofunc) 0,
  (setattrofunc) 0,
  
  (PyBufferProcs *) 0,
  0,
  /* documentation string */
  "This is a Python wrapper for a sidl object or interface."
};

static struct PyMethodDef spoa_methods[] = {
  /* this module exports no methods */
  { NULL, NULL }
};

#ifdef __cplusplus
extern "C" void initsidlObjA(void);
#else
extern void initsidlObjA(void);
#endif

void
initsidlObjA(void)
{
  PyObject *module, *dict, *c_api;
  static void *spoa_api[sidl_API_pointers];

  sidlPythonObjectType.ob_type = &PyType_Type;

  module = Py_InitModule("sidlObjA", spoa_methods);
  dict = PyModule_GetDict(module);
  spoa_api[sidl_Object_Create_NUM] = (void *)sidl_Object_Create;
  spoa_api[sidl_Get_IOR_NUM] = (void *)sidl_Get_IOR;
  spoa_api[sidl_Cast_NUM] = (void *)sidl_Cast;
  spoa_api[sidl_Interface_Create_NUM] = (void *)sidl_Interface_Create;
  spoa_api[sidl_Interface_Check_NUM] = (void *)sidl_Interface_Check;
  spoa_api[sidl_Opaque_Create_NUM] = (void *)sidl_Opaque_Create;
  spoa_api[sidl_Opaque_Convert_NUM] = (void *)sidl_Opaque_Convert;
  spoa_api[sidl_PyExceptionCast_NUM] = (void *)sidl_PyExceptionCast;
  spoa_api[sidl_PyType_NUM] = (void *)sidl_PyType;
  c_api = PyCObject_FromVoidPtr((void *)spoa_api, NULL);
  if (c_api) {
    PyDict_SetItemString(dict, "_C_API", c_api);
    Py_DECREF(c_api);
  }
  if (PyErr_Occurred()) {
    Py_FatalError("Can't initialize module sidlObjA.");
  }
}
